#include <stdint.h>

#include "log.h"

#include "nmea.h"
#include "gps.h"
#include "mstimer.h"
#include "clk.h"
#include "clkutc.h"
#include "clktime.h"

time64 NmeaTime = 0;
static uint32_t msTimerRepetitive = 0;

bool NmeaTimeAllowMessages = false; //This is set by gps from pps. 

int NmeaTimeMs()
{
    return MsTimerCount - msTimerRepetitive;
}

void NmeaTimeAlign()
{
    if (!NmeaTimeAllowMessages) return;
    if (msTimerRepetitive)
    {
        uint32_t elapsed = MsTimerCount - msTimerRepetitive;
        if (elapsed > 500) NmeaTime++;
    }
    msTimerRepetitive = MsTimerCount;
}

static void incrementTimeEverySecond()
{
    if (msTimerRepetitive && MsTimerRepetitive(&msTimerRepetitive, 1000)) NmeaTime++;
}

#define MIN_MS_SINCE_PULSE 100
#define MAX_MS_SINCE_PULSE 900
#define TIMEOUT_MS        3000
#define CONFIDENCE_OK_CLK_AGREES         10
#define CONFIDENCE_OK_CLK_DISAGREES 15 * 60

static int confidence      =  0;
static int confidenceRequirement = -1;
bool NmeaTimeIsStable() { return confidence == confidenceRequirement; }

static uint32_t msTimerMsgTimeout = (uint32_t)-TIMEOUT_MS;
static bool nmeaWasStopped = true;

void NmeaTimeReceived(time64 nmeaMsgTime, int nmeaMs, int nmeaStartMs)
{    
    //Ignore messages until the time is aligned
    if (!msTimerRepetitive) return;
    
    //Reset the msg timeout timer
    msTimerMsgTimeout = MsTimerCount;
    if (nmeaWasStopped) GpsLog("NMEA time messages have started\r\n");
    nmeaWasStopped = false;
    
    //Ignore time messages when the PPS is not running
    if (!NmeaTimeAllowMessages) return;
    
    //Ignore time messages with non zero ms in NMEA
    if (nmeaMs)
    {
        GpsLog("NMEA time has non zero ms = %d\r\n", nmeaMs);
        return;
    }

    //Add time to bins to allow display of distribution
    NmeaTimeBucketAdd(nmeaStartMs);

    //Ignore time messages which have probably overflowed from a previous pulse
    if (nmeaStartMs < MIN_MS_SINCE_PULSE)
    {
        GpsLogVerbose("NMEA time message ignored as start time %d ms is before %d ms\r\n", nmeaStartMs, MIN_MS_SINCE_PULSE);
        return;
    }
    
    //Ignore time messages which have probably overflowed from a previous pulse
    if (nmeaStartMs > MAX_MS_SINCE_PULSE)
    {
        GpsLogVerbose("NMEA time message ignored as start time %d ms is after %d ms\r\n", nmeaStartMs, MAX_MS_SINCE_PULSE);
        return;
    }
    
    //Set up the confidence requirement
    static int confidenceRequirementLast = 0;
    time64 clkTime = ClkUtcFromTai(ClkNowTai()) >> CLK_TIME_ONE_SECOND_SHIFT;
    if (nmeaMsgTime == clkTime)
    {
        confidenceRequirement = CONFIDENCE_OK_CLK_AGREES;
        if (confidenceRequirement != confidenceRequirementLast)
        {
            GpsLog("NMEA time in sentence and clock time agree so set confidence requirement to %d\r\n", confidenceRequirement);
        }
    }
    else
    {
        confidenceRequirement = CONFIDENCE_OK_CLK_DISAGREES;
        if (confidenceRequirement != confidenceRequirementLast)
        {
            GpsLog("NMEA time in sentence and clock time disagree so set confidence requirement to %d\r\n", confidenceRequirement);
        }
    }
    confidenceRequirementLast = confidenceRequirement;
    
    //Handle the time message
    if (NmeaTime == nmeaMsgTime) 
    {
        confidence++;
        if (confidence > confidenceRequirement) confidence = confidenceRequirement;
    }
    else
    {
        confidence = 0;
        if (NmeaTime) GpsLog("NMEA time differs from time in sentence by %+d seconds so setting to sentence - sync disabled\r\n", nmeaMsgTime - NmeaTime);
        else          GpsLog("NMEA time not set so setting to time in sentence - sync disabled\r\n");
        NmeaTime = nmeaMsgTime;
    }
}
static void checkNmeaIsReady()
{
    //Disable syncing until Nmea module is ready
    if (NmeaModuleReadiness != NMEA_READINESS_READY) confidence = 0;
}
static void checkIfReceivingNmeaTime()
{
    if (MsTimerRelative(msTimerMsgTimeout, TIMEOUT_MS))
    {
        confidence = 0;
        if (!nmeaWasStopped) GpsLog("NMEA time messages have stopped - sync disabled\r\n");
        nmeaWasStopped = true;
    }    
}
static void buildConfidence()
{
    static bool wasConfident = false;
    if (confidence == confidenceRequirement)
    {
        if (!wasConfident) GpsLog("NMEA time is stable\r\n");
        wasConfident = true;
    }
    else
    {
        if (wasConfident) GpsLog("NMEA time is not stable\r\n");
        wasConfident = false;
    }

}
void NmeaTimeMain()
{
    checkIfReceivingNmeaTime();
    buildConfidence();
    checkNmeaIsReady();
    incrementTimeEverySecond();
}
