#include <stdbool.h>
#include <time.h>

#include "log.h"
#include "clkgov.h"
#include "hrtimer.h"
#include "gpio.h"
#include "bitband.h"
#include "gps.h"

#define IO0INTENR_ADDR 0x40028090
#define IO0INTCLR_ADDR 0x4002808C
#define     ISER0_ADDR 0xE000E100

#define     FIX_PIN FIO0PIN(23)
#define  ENABLE_DIR FIO0DIR(24)
#define  ENABLE_SET FIO0SET(24)
#define  ENABLE_CLR FIO0CLR(24)
#define     PPS_PIN FIO0PIN(17)
#define PPS_INT_ENR BIT_BAND4(IO0INTENR_ADDR, 17) = 1
#define PPS_INT_CLR BIT_BAND4(IO0INTCLR_ADDR, 17) = 1
#define ISER0 *((volatile unsigned *) ISER0_ADDR)

static volatile uint32_t hrThisPulse = 0;
static volatile int hadPulse = false;

#define CONFIDENCE_OK  10
#define TIMEOUT_MS   1000

static int confidence = 0;
bool PpsIsStable() { return confidence == CONFIDENCE_OK; }

int  PpsMsSinceLastPulse = 0;

void PpsHandler() //Called by an interrupt
{
    PPS_INT_CLR;
    ClkGovSyncPpsI();
    hrThisPulse = HrTimerCount();
    hadPulse = true;
}
static bool pulseWasStopped = true;
static void checkForPulsesStoppingAndStarting()
{
    if (PpsMsSinceLastPulse > TIMEOUT_MS)
    {
        confidence = 0;
        if (!pulseWasStopped)
        {
            GpsLog("PPS pulses have stopped - sync disabled\r\n");
            pulseWasStopped = true;
        }
    }
    if (hadPulse)
    {
        if (pulseWasStopped)
        {
            GpsLog("PPS pulses have started\r\n");
            pulseWasStopped = false;
        }
    }
}

static void buildConfidence()
{
    static uint32_t hrPrevPulse = 0;
    if (hrPrevPulse)
    {
        uint32_t hrSinceLastPulse        = hrThisPulse - hrPrevPulse;
        uint32_t msSinceLastPulse        = hrSinceLastPulse / (HR_TIMER_COUNT_PER_SECOND / 1000);
        uint32_t msSinceLastPulseMod1000 = msSinceLastPulse % 1000;

        if (msSinceLastPulseMod1000 == 999 || msSinceLastPulseMod1000 == 000)
        {
            confidence++;
            if (confidence > CONFIDENCE_OK) confidence = CONFIDENCE_OK;
        }
        else
        {
            GpsLog("PPS interval %03d is not 999 or 1000 ms - sync disabled\r\n", msSinceLastPulse);
            confidence = 0;
        }
    }
    hrPrevPulse = hrThisPulse;
}
static void displayConfidence()
{
    static bool wasConfident = false;
    if (confidence == CONFIDENCE_OK)
    {
        if (!wasConfident) GpsLog("PPS is stable\r\n");
        wasConfident = true;
    }
    else
    {
        if (wasConfident) GpsLog("PPS is not stable\r\n");
        wasConfident = false;
    }
}

void PpsMain()
{
    //Wait for 100ms after the start up in order to ignore initial glitches on the PPS line
    static bool afterStartup = false; //Need to store the timeout as hr timer rolls over every 44 seconds
    if (HrTimerCount() > HR_TIMER_COUNT_PER_SECOND / 10) afterStartup = true;
    if (!afterStartup)
    {
        hadPulse = false; //Cancel the pulse
        return;
    }
    
    checkForPulsesStoppingAndStarting(); //Do this before storing PPsMsSinceLastPulse to allow log to show when they started not zero
    
    PpsMsSinceLastPulse = HrTimerSinceMs(hrThisPulse);
    
    if (hadPulse)
    {
        buildConfidence();
        GpsHadPps();
    }
    displayConfidence();

    hadPulse = false;
}

void PpsInit()
{
    ENABLE_SET;       //Set the enable to high
    ENABLE_DIR = 1;   //Set the enable pin direction to 1 == output
    PPS_INT_ENR;      //Set the PPS pin to be interrupt on rise
    ISER0 |= 1 << 21; //6.5.1 bit1 == Interrupt set enable for EINT3. It MUST be enabled even for GPIO interrupts - I checked.
}